@@ -1,11 +1,9 @@
 package Bitcoin::Crypto::BIP44;
-$Bitcoin::Crypto::BIP44::VERSION = '4.002';
-use v5.10;
-use strict;
+$Bitcoin::Crypto::BIP44::VERSION = '4.003';
+use v5.14;
 use warnings;
-use Moo;
-use Mooish::AttributeBuilder -standard;
-use Types::Common -sigs, -types;
+
+use Mooish::Base -standard;
 
 use Scalar::Util qw(blessed);
 
@@ -14,8 +12,6 @@ use Bitcoin::Crypto::Types -types;
 use Bitcoin::Crypto::Network;
 use Bitcoin::Crypto::Exception;
 
-use namespace::clean;
-
 sub _get_network_constant
 {
 	my ($network) = @_;
@@ -92,11 +88,6 @@ use overload
 	q{""} => sub { shift->as_string },
 	fallback => 1;
 
-signature_for as_string => (
-	method => Object,
-	positional => [],
-);
-
 sub as_string
 {
 	my ($self) = @_;
@@ -117,11 +108,6 @@ sub as_string
 		$path, $self->change, $self->index;
 }
 
-signature_for get_derivation_path => (
-	method => Object,
-	positional => [],
-);
-
 sub get_derivation_path
 {
 	my ($self) = @_;
@@ -269,18 +255,6 @@ indirectly by just stringifying the object.
 
 Returns an object of class L<Bitcoin::Crypto::DerivationPath>.
 
-=head1 EXCEPTIONS
-
-This module throws an instance of L<Bitcoin::Crypto::Exception> if it
-encounters an error. It can produce the following error types from the
-L<Bitcoin::Crypto::Exception> namespace:
-
-=over
-
-=item * NetworkConfig - incomplete or corrupted network configuration
-
-=back
-
 =head1 SEE ALSO
 
 L<Bitcoin::Crypto::Key::ExtPrivate>
