@@ -1,26 +1,23 @@
 package Bitcoin::Crypto::Key::ExtBase;
-$Bitcoin::Crypto::Key::ExtBase::VERSION = '4.002';
-use v5.10;
-use strict;
+$Bitcoin::Crypto::Key::ExtBase::VERSION = '4.003';
+use v5.14;
 use warnings;
-use Moo;
+
+use Mooish::Base -standard;
 use Scalar::Util qw(blessed);
-use Mooish::AttributeBuilder -standard;
-use Types::Common -sigs, -types;
+use Types::Common -sigs;
 use Carp qw(carp);
 use List::Util qw(none);
 
 use Bitcoin::Crypto::Key::Private;
 use Bitcoin::Crypto::Key::Public;
-use Bitcoin::Crypto::Constants;
+use Bitcoin::Crypto::Constants qw(:bip44 :key);
 use Bitcoin::Crypto::Types -types;
-use Bitcoin::Crypto::Util qw(hash160 to_format);
+use Bitcoin::Crypto::Util::Internal qw(hash160 to_format);
 use Bitcoin::Crypto::Helpers qw(ensure_length);
 use Bitcoin::Crypto::Network;
 use Bitcoin::Crypto::Exception;
 
-use namespace::clean;
-
 has param 'depth' => (
 	isa => IntMaxBits [8],
 	default => 0
@@ -60,18 +57,13 @@ sub _get_network_extkey_version
 
 	my $name = 'ext';
 	$name .= $self->_is_private ? 'prv' : 'pub';
-	$name .= '_compat' if $purpose && $purpose eq Bitcoin::Crypto::Constants::bip44_compat_purpose;
-	$name .= '_segwit' if $purpose && $purpose eq Bitcoin::Crypto::Constants::bip44_segwit_purpose;
+	$name .= '_compat' if $purpose && $purpose eq BIP44_COMPAT_PURPOSE;
+	$name .= '_segwit' if $purpose && $purpose eq BIP44_SEGWIT_PURPOSE;
 	$name .= '_version';
 
 	return $network->$name;
 }
 
-signature_for to_serialized => (
-	method => Object,
-	positional => [],
-);
-
 sub to_serialized
 {
 	my ($self) = @_;
@@ -99,13 +91,13 @@ sub to_serialized
 	$serialized .= $self->chain_code;
 
 	# key entropy (1 + 32B)
-	$serialized .= ensure_length $self->raw_key, Bitcoin::Crypto::Constants::key_max_length + 1;
+	$serialized .= ensure_length $self->raw_key, KEY_MAX_LENGTH + 1;
 
 	return $serialized;
 }
 
 signature_for from_serialized => (
-	method => Str,
+	method => !!1,
 	positional => [ByteStr, Maybe [Str], {default => undef}],
 );
 
@@ -115,9 +107,8 @@ sub from_serialized
 
 	# expected length is 78
 	if (defined $serialized && length $serialized == 78) {
-		my $format = 'a4aa4a4a32a33';
 		my ($version, $depth, $fingerprint, $number, $chain_code, $data) =
-			unpack($format, $serialized);
+			unpack 'a4aa4a4a32a33', $serialized;
 
 		my $is_private = pack('x') eq substr $data, 0, 1;
 
@@ -125,7 +116,7 @@ sub from_serialized
 			'invalid class used, key is ' . ($is_private ? 'private' : 'public')
 		) if $is_private != $class->_is_private;
 
-		$data = substr $data, 1, Bitcoin::Crypto::Constants::key_max_length
+		$data = substr $data, 1, KEY_MAX_LENGTH
 			if $is_private;
 
 		$version = unpack 'N', $version;
@@ -135,8 +126,8 @@ sub from_serialized
 
 		for my $check_purpose (
 			undef,
-			Bitcoin::Crypto::Constants::bip44_compat_purpose,
-			Bitcoin::Crypto::Constants::bip44_segwit_purpose
+			BIP44_COMPAT_PURPOSE,
+			BIP44_SEGWIT_PURPOSE
 			)
 		{
 			@found_networks = Bitcoin::Crypto::Network->find(
@@ -175,7 +166,7 @@ sub from_serialized
 		) if @found_networks == 0;
 
 		my $key = $class->new(
-			key_instance => $data,
+			_key_instance => $data,
 			chain_code => $chain_code,
 			child_number => unpack('N', $number),
 			parent_fingerprint => $fingerprint,
@@ -193,17 +184,12 @@ sub from_serialized
 	}
 }
 
-signature_for get_basic_key => (
-	method => Object,
-	positional => [],
-);
-
 sub get_basic_key
 {
 	my ($self) = @_;
 	my $base_class = 'Bitcoin::Crypto::Key::' . ($self->_is_private ? 'Private' : 'Public');
 	my $basic_key = $base_class->new(
-		key_instance => $self->key_instance,
+		_key_instance => $self->_key_instance,
 		network => $self->network,
 		purpose => $self->purpose,
 	);
@@ -212,7 +198,7 @@ sub get_basic_key
 }
 
 signature_for get_fingerprint => (
-	method => Object,
+	method => !!1,
 	positional => [PositiveInt, {default => 4}],
 );
 
@@ -245,7 +231,7 @@ sub _get_purpose_from_BIP44
 }
 
 signature_for derive_key => (
-	method => Object,
+	method => !!1,
 	positional => [Defined],
 );
 
@@ -262,7 +248,7 @@ sub derive_key
 
 	my $key = $self;
 	for my $child_num (@{$path_info->path}) {
-		my $hardened = $child_num >= Bitcoin::Crypto::Constants::max_child_keys;
+		my $hardened = $child_num >= MAX_CHILD_KEYS;
 
 		# dies if hardened-from-public requested
 		# dies if key is invalid
