@@ -1,36 +1,55 @@
 package Bitcoin::Crypto::Key::ExtPrivate;
-$Bitcoin::Crypto::Key::ExtPrivate::VERSION = '4.002';
-use v5.10;
-use strict;
+$Bitcoin::Crypto::Key::ExtPrivate::VERSION = '4.003';
+use v5.14;
 use warnings;
-use Moo;
+
+use Mooish::Base -standard;
 use Crypt::Mac::HMAC qw(hmac);
 use Bitcoin::BIP39 qw(bip39_mnemonic_to_entropy);
-use Types::Common -sigs, -types;
+use Types::Common -sigs;
 use Carp qw(carp);
 
 use Bitcoin::Crypto::BIP44;
 use Bitcoin::Crypto::Key::ExtPublic;
 use Bitcoin::Crypto::Constants;
-use Bitcoin::Crypto::Helpers qw(ensure_length ecc);
-use Bitcoin::Crypto::Util qw(mnemonic_to_seed);
+use Bitcoin::Crypto::Helpers qw(ensure_length ecc die_no_trace);
+use Bitcoin::Crypto::Util::Internal qw(mnemonic_to_seed);
 use Bitcoin::Crypto::Types -types;
 use Bitcoin::Crypto::Exception;
 
-use namespace::clean;
-
 extends qw(Bitcoin::Crypto::Key::ExtBase);
 
 sub _is_private { 1 }
 
+signature_for from_serialized => (
+	method => !!1,
+	positional => [BitcoinSecret, Maybe [Str], {default => undef}],
+);
+
+sub from_serialized
+{
+	my ($class, $secret, $network) = @_;
+
+	return $secret->unmask_to(
+		sub {
+			return $class->SUPER::from_serialized(shift, $network);
+		}
+	);
+}
+
 signature_for from_mnemonic => (
-	method => Str,
-	positional => [Str, Maybe [Str], {default => ''}, Maybe [Str], {default => undef}],
+	method => !!1,
+	positional => [BitcoinSecret, Maybe [BitcoinSecret], {default => ''}, Maybe [Str], {default => undef}],
 );
 
 sub from_mnemonic
 {
-	my ($class, $mnemonic, $password, $lang) = @_;
+	my ($class, $secret_mnemonic, $secret_password, $lang) = @_;
+
+	my $mnemonic = $secret_mnemonic->unmask_to(sub { shift });
+	my $password = defined $secret_password
+		? $secret_password->unmask_to(sub { shift })
+		: undef;
 
 	if (defined $lang) {
 
@@ -54,35 +73,37 @@ sub from_mnemonic
 }
 
 signature_for from_seed => (
-	method => Str,
-	positional => [ByteStr],
+	method => !!1,
+	positional => [BitcoinSecret],
 );
 
 sub from_seed
 {
-	my ($class, $seed) = @_;
+	my ($class, $secret) = @_;
+	state $sig = signature(positional => [ByteStr]);
 
-	my $bytes = hmac('SHA512', 'Bitcoin seed', $seed);
-	my $key = substr $bytes, 0, 32;
-	my $cc = substr $bytes, 32, 32;
+	return $secret->unmask_to(
+		sub {
+			my ($seed) = $sig->(@_);
 
-	return $class->new(
-		key_instance => $key,
-		chain_code => $cc,
+			my $bytes = hmac('SHA512', 'Bitcoin seed', $seed);
+			my $key = substr $bytes, 0, 32;
+			my $cc = substr $bytes, 32, 32;
+
+			return $class->new(
+				_key_instance => $key,
+				chain_code => $cc,
+			);
+		}
 	);
 }
 
-signature_for get_public_key => (
-	method => Object,
-	positional => [],
-);
-
 sub get_public_key
 {
 	my ($self) = @_;
 
 	my $public = Bitcoin::Crypto::Key::ExtPublic->new(
-		key_instance => $self->raw_key('public'),
+		_key_instance => $self->raw_key('public'),
 		chain_code => $self->chain_code,
 		child_number => $self->child_number,
 		parent_fingerprint => $self->parent_fingerprint,
@@ -94,16 +115,11 @@ sub get_public_key
 	return $public;
 }
 
-signature_for derive_key_bip44 => (
-	method => Object,
-	positional => [HashRef, {slurpy => !!1}],
-);
-
 sub derive_key_bip44
 {
-	my ($self, $data) = @_;
+	my ($self, %data) = @_;
 	my $path = Bitcoin::Crypto::BIP44->new(
-		%{$data},
+		%data,
 		coin_type => $self,
 	);
 
@@ -137,13 +153,13 @@ sub _derive_key_partial
 	Bitcoin::Crypto::Exception::KeyDerive->trap_into(
 		sub {
 			$key = ecc->add_private_key($key, $tweak);
-			die 'verification failed' unless ecc->verify_private_key($key);
+			die_no_trace 'verification failed' unless ecc->verify_private_key($key);
 		},
 		"key $child_num in sequence was found invalid"
 	);
 
 	return $self->new(
-		key_instance => $key,
+		_key_instance => $key,
 		chain_code => $chain_code,
 		child_number => $child_num,
 		parent_fingerprint => $self->get_fingerprint,
@@ -263,6 +279,8 @@ produce a different key>. Be careful when using this method without C<$lang>
 argument as you can easily create keys incompatible with other software due to
 these whitespace problems.
 
+This method accepts a secret argument. See L<Bitcoin::Crypto::Secret> for details.
+
 Returns a new instance of this class.
 
 B<Important note about unicode:> this function only accepts UTF8-decoded
@@ -279,6 +297,8 @@ wallet.
 Creates and returns a new key from seed, which can be any data of any length.
 C<$seed> is expected to be a byte string.
 
+This method accepts a secret argument. See L<Bitcoin::Crypto::Secret> for details.
+
 =head3 to_serialized
 
 	$serialized = $object->to_serialized()
@@ -294,6 +314,8 @@ Tries to unserialize byte string C<$serialized> with format specified in BIP32.
 Dies on errors. If multiple networks match serialized data specify C<$network>
 manually (id of the network) to avoid exception.
 
+This method accepts a secret argument. See L<Bitcoin::Crypto::Secret> for details.
+
 =head3 set_network
 
 	$object->set_network($val)
@@ -347,26 +369,6 @@ configuration set in the extended key will be used.>
 
 Returns a fingerprint of the extended key of C<$len> length (byte string)
 
-=head1 EXCEPTIONS
-
-This module throws an instance of L<Bitcoin::Crypto::Exception> if it
-encounters an error. It can produce the following error types from the
-L<Bitcoin::Crypto::Exception> namespace:
-
-=over
-
-=item * MnemonicGenerate - mnemonic couldn't be generated correctly
-
-=item * MnemonicCheck - mnemonic didn't pass the validity check
-
-=item * KeyDerive - key couldn't be derived correctly
-
-=item * KeyCreate - key couldn't be created correctly
-
-=item * NetworkConfig - incomplete or corrupted network configuration
-
-=back
-
 =head1 SEE ALSO
 
 =over
