@@ -1,18 +1,13 @@
 package Bitcoin::Crypto::Network;
-$Bitcoin::Crypto::Network::VERSION = '4.002';
-use v5.10;
-use strict;
+$Bitcoin::Crypto::Network::VERSION = '4.003';
+use v5.14;
 use warnings;
 
-use Moo;
-use Mooish::AttributeBuilder -standard;
-use Types::Common -sigs, -types;
+use Mooish::Base -standard;
 
 use Bitcoin::Crypto::Exception;
 use Bitcoin::Crypto::Types -types;
 
-use namespace::clean;
-
 my %networks;
 my $default_network;
 my $single_network = 0;
@@ -78,11 +73,6 @@ has param 'bip44_coin' => (
 	required => 0,
 );
 
-signature_for single_network => (
-	method => Str,
-	positional => [],
-);
-
 sub single_network
 {
 	my ($class) = @_;
@@ -90,28 +80,18 @@ sub single_network
 	return $single_network;
 }
 
-signature_for register => (
-	method => !!1,
-	positional => [HashRef, {slurpy => !!1}],
-);
-
 sub register
 {
-	my ($self, $config) = @_;
+	my ($self, %config) = @_;
 
 	if (!ref $self) {
-		$self = $self->new($config);
+		$self = $self->new(%config);
 	}
 
 	$networks{$self->id} = $self;
 	return $self;
 }
 
-signature_for unregister => (
-	method => Object,
-	positional => [],
-);
-
 sub unregister
 {
 	my ($self) = @_;
@@ -124,11 +104,6 @@ sub unregister
 	return $self;
 }
 
-signature_for set_default => (
-	method => Object,
-	positional => [],
-);
-
 sub set_default
 {
 	my ($self) = @_;
@@ -142,11 +117,6 @@ sub set_default
 	return $self;
 }
 
-signature_for set_single => (
-	method => Object,
-	positional => [],
-);
-
 sub set_single
 {
 	my ($self) = @_;
@@ -157,11 +127,6 @@ sub set_single
 	return $self;
 }
 
-signature_for supports_segwit => (
-	method => Object,
-	positional => [],
-);
-
 sub supports_segwit
 {
 	my ($self) = @_;
@@ -169,11 +134,6 @@ sub supports_segwit
 	return defined $self->segwit_hrp;
 }
 
-signature_for find => (
-	method => Str,
-	positional => [Maybe [CodeRef], {default => undef}],
-);
-
 sub find
 {
 	my ($class, $sub) = @_;
@@ -184,21 +144,13 @@ sub find
 	return grep { $sub->($networks{$_}) } keys %networks;
 }
 
-signature_for get => (
-	method => Str,
-	positional => [Str, {default => sub { $default_network }}],
-);
-
 sub get
 {
-	my ($class, $id) = @_;
-
-	my $network = $networks{$id};
-	Bitcoin::Crypto::Exception::NetworkConfig->raise(
-		"network $id is not registered"
-	) unless defined $network;
-
-	return $network;
+	my $id = $_[1] // $default_network;
+	return $networks{$id}
+		// Bitcoin::Crypto::Exception::NetworkConfig->raise(
+			"network $id is not registered"
+		);
 }
 
 ### PREDEFINED NETWORKS SECTION
