@@ -1,12 +1,10 @@
 package Bitcoin::Crypto::PSBT;
-$Bitcoin::Crypto::PSBT::VERSION = '4.002';
-use v5.10;
-use strict;
+$Bitcoin::Crypto::PSBT::VERSION = '4.003';
+use v5.14;
 use warnings;
 
-use Moo;
-use Mooish::AttributeBuilder -standard;
-use Types::Common -sigs, -types;
+use Mooish::Base -standard;
+use Types::Common -sigs;
 use List::Util qw(any);
 use Scalar::Util qw(blessed);
 
@@ -14,11 +12,9 @@ use Bitcoin::Crypto::PSBT::Map;
 use Bitcoin::Crypto::PSBT::Field;
 use Bitcoin::Crypto::PSBT::FieldType;
 use Bitcoin::Crypto::Types -types;
-use Bitcoin::Crypto::Constants;
+use Bitcoin::Crypto::Constants qw(:psbt);
 use Bitcoin::Crypto::Exception;
 
-use namespace::clean;
-
 has field 'maps' => (
 	isa => ArrayRef [InstanceOf ['Bitcoin::Crypto::PSBT::Map']],
 	default => sub { [] },
@@ -49,11 +45,6 @@ sub _get_map
 	return $found_map;
 }
 
-signature_for input_count => (
-	method => Object,
-	positional => []
-);
-
 sub input_count
 {
 	my ($self) = @_;
@@ -68,11 +59,6 @@ sub input_count
 	}
 }
 
-signature_for output_count => (
-	method => Object,
-	positional => []
-);
-
 sub output_count
 {
 	my ($self) = @_;
@@ -88,7 +74,7 @@ sub output_count
 }
 
 signature_for get_field => (
-	method => Object,
+	method => !!1,
 	positional => [
 		PSBTFieldType,
 		Maybe [PositiveOrZeroInt], {default => undef},
@@ -109,7 +95,7 @@ sub get_field
 }
 
 signature_for get_all_fields => (
-	method => Object,
+	method => !!1,
 	positional => [
 		PSBTFieldType,
 		Maybe [PositiveOrZeroInt], {default => undef},
@@ -126,22 +112,17 @@ sub get_all_fields
 	return $map->find($type, $key);
 }
 
-signature_for add_field => (
-	method => Object,
-	positional => [ArrayRef, {slurpy => !!1}],
-);
-
 sub add_field
 {
-	my ($self, $data) = @_;
+	my ($self, @data) = @_;
 	my $field;
 	my $index;
 
-	if ((@$data == 1 || @$data == 2) && blessed $data->[0] && $data->[0]->isa('Bitcoin::Crypto::PSBT::Field')) {
-		($field, $index) = @$data;
+	if ((@data == 1 || @data == 2) && blessed $data[0] && $data[0]->isa('Bitcoin::Crypto::PSBT::Field')) {
+		($field, $index) = @data;
 	}
 	else {
-		my %data = @$data;
+		my %data = @data;
 		$index = delete $data{index};
 		$field = Bitcoin::Crypto::PSBT::Field->new(%data);
 	}
@@ -152,11 +133,6 @@ sub add_field
 	return $self;
 }
 
-signature_for list_fields => (
-	method => Object,
-	positional => [],
-);
-
 sub list_fields
 {
 	my ($self) = @_;
@@ -177,11 +153,6 @@ sub list_fields
 	return @results[0 .. $#results];
 }
 
-signature_for version => (
-	method => Object,
-	positional => [],
-);
-
 sub version
 {
 	my ($self) = @_;
@@ -197,7 +168,7 @@ sub version
 }
 
 signature_for from_serialized => (
-	method => Str,
+	method => !!1,
 	positional => [ByteStr],
 );
 
@@ -206,23 +177,23 @@ sub from_serialized
 	my ($class, $serialized) = @_;
 	my $self = $class->new;
 
-	my $pos = length Bitcoin::Crypto::Constants::psbt_magic;
+	my $pos = length PSBT_MAGIC;
 	my $magic = substr $serialized, 0, $pos;
 
 	Bitcoin::Crypto::Exception::PSBT->raise(
 		'serialized string does not contain the PSBT header'
-	) unless $magic eq Bitcoin::Crypto::Constants::psbt_magic;
+	) unless $magic eq PSBT_MAGIC;
 
 	push @{$self->maps}, Bitcoin::Crypto::PSBT::Map::->from_serialized(
 		$serialized,
-		map_type => Bitcoin::Crypto::Constants::psbt_global_map,
+		map_type => PSBT_GLOBAL_MAP,
 		pos => \$pos,
 	);
 
 	foreach my $index (0 .. $self->input_count - 1) {
 		push @{$self->maps}, Bitcoin::Crypto::PSBT::Map::->from_serialized(
 			$serialized,
-			map_type => Bitcoin::Crypto::Constants::psbt_input_map,
+			map_type => PSBT_INPUT_MAP,
 			pos => \$pos,
 			index => $index,
 		);
@@ -231,7 +202,7 @@ sub from_serialized
 	foreach my $index (0 .. $self->output_count - 1) {
 		push @{$self->maps}, Bitcoin::Crypto::PSBT::Map::->from_serialized(
 			$serialized,
-			map_type => Bitcoin::Crypto::Constants::psbt_output_map,
+			map_type => PSBT_OUTPUT_MAP,
 			pos => \$pos,
 			index => $index,
 		);
@@ -246,38 +217,28 @@ sub from_serialized
 	return $self;
 }
 
-signature_for to_serialized => (
-	method => Object,
-	positional => [],
-);
-
 sub to_serialized
 {
 	my ($self) = @_;
 
 	$self->check;
 
-	my $serialized = Bitcoin::Crypto::Constants::psbt_magic;
-	$serialized .= $self->_get_map(Bitcoin::Crypto::Constants::psbt_global_map, set => !!1)->to_serialized;
+	my $serialized = PSBT_MAGIC;
+	$serialized .= $self->_get_map(PSBT_GLOBAL_MAP, set => !!1)->to_serialized;
 
 	for my $input_index (0 .. $self->input_count - 1) {
-		$serialized .= $self->_get_map(Bitcoin::Crypto::Constants::psbt_input_map, index => $input_index, set => !!1)
+		$serialized .= $self->_get_map(PSBT_INPUT_MAP, index => $input_index, set => !!1)
 			->to_serialized;
 	}
 
 	for my $output_index (0 .. $self->output_count - 1) {
-		$serialized .= $self->_get_map(Bitcoin::Crypto::Constants::psbt_output_map, index => $output_index, set => !!1)
+		$serialized .= $self->_get_map(PSBT_OUTPUT_MAP, index => $output_index, set => !!1)
 			->to_serialized;
 	}
 
 	return $serialized;
 }
 
-signature_for check => (
-	method => Object,
-	positional => [],
-);
-
 sub check
 {
 	my ($self) = @_;
@@ -289,13 +250,13 @@ sub check
 	foreach my $field_type (@{$required_fields}) {
 		my @maps;
 
-		if ($field_type->map_type eq Bitcoin::Crypto::Constants::psbt_global_map) {
+		if ($field_type->map_type eq PSBT_GLOBAL_MAP) {
 			@maps = ($self->_get_map($field_type->map_type));
 		}
-		elsif ($field_type->map_type eq Bitcoin::Crypto::Constants::psbt_input_map) {
+		elsif ($field_type->map_type eq PSBT_INPUT_MAP) {
 			@maps = map { $self->_get_map($field_type->map_type, index => $_) } 0 .. $input_count - 1;
 		}
-		elsif ($field_type->map_type eq Bitcoin::Crypto::Constants::psbt_output_map) {
+		elsif ($field_type->map_type eq PSBT_OUTPUT_MAP) {
 			@maps = map { $self->_get_map($field_type->map_type, index => $_) } 0 .. $output_count - 1;
 		}
 
@@ -309,12 +270,12 @@ sub check
 
 	foreach my $map (@{$self->maps}) {
 
-		if ($map->type eq Bitcoin::Crypto::Constants::psbt_input_map) {
+		if ($map->type eq PSBT_INPUT_MAP) {
 			Bitcoin::Crypto::Exception::PSBT->raise(
 				"PSBT input map index " . $map->index . " out of range"
 			) unless $map->index < $input_count;
 		}
-		elsif ($map->type eq Bitcoin::Crypto::Constants::psbt_output_map) {
+		elsif ($map->type eq PSBT_OUTPUT_MAP) {
 			Bitcoin::Crypto::Exception::PSBT->raise(
 				"PSBT output map index " . $map->index . " out of range"
 			) unless $map->index < $output_count;
@@ -330,11 +291,6 @@ sub check
 	return $self;
 }
 
-signature_for dump => (
-	method => Object,
-	positional => [],
-);
-
 sub dump
 {
 	my ($self) = @_;
@@ -530,7 +486,7 @@ Returns a readable description of all the maps in the PSBT.
 Just for reference, here is a list of all PSBT fields which are currently
 supported by this module. This list is auto-generated from
 L<Bitcoin::Crypto::PSBT::FieldType/key_data> and
-C<Bitcoin::Crypto::PSBT::FieldType/value_data> (which contain short strings
+L<Bitcoin::Crypto::PSBT::FieldType/value_data> (which contain short strings
 with description of field content).
 
 For a list of PSBT fields, see
@@ -766,7 +722,7 @@ B<Value data:> Array reference, where the first item is a fingerprint and the se
 
 B<Key data:> <none>
 
-B<Value data:> Math::BigInt object
+B<Value data:> 64 bit number
 
 =item * PSBT_OUT_SCRIPT
 
@@ -794,18 +750,6 @@ B<Value data:> Array reference, where first item is an array of leaf hashes, sec
 
 =back
 
-=head1 EXCEPTIONS
-
-This module throws an instance of L<Bitcoin::Crypto::Exception> if it
-encounters an error. It can produce the following error types from the
-L<Bitcoin::Crypto::Exception> namespace:
-
-=over
-
-=item * PSBT - general error with the PSBT
-
-=back
-
 =head1 SEE ALSO
 
 =over
