@@ -1,26 +1,22 @@
 package Bitcoin::Crypto::PSBT::FieldType;
-$Bitcoin::Crypto::PSBT::FieldType::VERSION = '4.002';
-use v5.10;
-use strict;
+$Bitcoin::Crypto::PSBT::FieldType::VERSION = '4.003';
+use v5.14;
 use warnings;
 
-use Moo;
-use Mooish::AttributeBuilder -standard;
-use Types::Common -sigs, -types;
+use Mooish::Base -standard;
+use Types::Common -sigs;
 use List::Util qw(any notall);
 
 use Bitcoin::Crypto qw(btc_extpub btc_pub btc_transaction btc_script btc_tapscript btc_script_tree);
 use Bitcoin::Crypto::Transaction::Output;
-use Bitcoin::Crypto::Constants;
+use Bitcoin::Crypto::Constants qw(:transaction);
 use Bitcoin::Crypto::Exception;
-use Bitcoin::Crypto::Util qw(pack_compactsize unpack_compactsize lift_x);
-use Bitcoin::Crypto::Helpers qw(ensure_length);    # loads Math::BigInt
+use Bitcoin::Crypto::Util::Internal qw(pack_compactsize unpack_compactsize lift_x);
+use Bitcoin::Crypto::Helpers qw(encode_64bit decode_64bit die_no_trace);
 use Bitcoin::Crypto::Types -types;
 use Bitcoin::Crypto::Transaction::ControlBlock;
 use Bitcoin::Crypto::DerivationPath;
 
-use namespace::clean;
-
 use constant {
 	REQUIRED => 'required',
 	AVAILABLE => 'available',
@@ -248,11 +244,11 @@ my %types = (
 		validator => sub {
 			my ($tx) = @_;
 
-			die 'must not have signatures'
+			die_no_trace 'must not have signatures'
 				if notall { $_->signature_script->is_empty } @{$tx->inputs};
 
-			die 'must be in non-witness format'
-				if any { $_->has_witness } @{$tx->inputs};
+			die_no_trace 'must be in non-witness format'
+				if $tx->had_witness_flag || any { $_->has_witness } @{$tx->inputs};
 		},
 		version_status => {
 			0 => REQUIRED,
@@ -531,8 +527,9 @@ my %types = (
 		%uint_32bitLE_serializers,
 		validator => sub {
 			my ($value) = @_;
-			die 'must be greather than or equal to ' . Bitcoin::Crypto::Constants::locktime_height_threshold
-				if $value < Bitcoin::Crypto::Constants::locktime_height_threshold;
+			die_no_trace 'must be greather than or equal to '
+				. LOCKTIME_HEIGHT_THRESHOLD
+				if $value < LOCKTIME_HEIGHT_THRESHOLD;
 		},
 		version_status => {
 			2 => AVAILABLE,
@@ -544,8 +541,8 @@ my %types = (
 		%uint_32bitLE_serializers,
 		validator => sub {
 			my ($value) = @_;
-			die 'must be less than ' . Bitcoin::Crypto::Constants::locktime_height_threshold
-				unless $value < Bitcoin::Crypto::Constants::locktime_height_threshold;
+			die_no_trace 'must be less than ' . LOCKTIME_HEIGHT_THRESHOLD
+				unless $value < LOCKTIME_HEIGHT_THRESHOLD;
 		},
 		version_status => {
 			2 => AVAILABLE,
@@ -558,7 +555,7 @@ my %types = (
 		validator => sub {
 			my ($value) = @_;
 			state $validator = ByteStrLen [64] | ByteStrLen [65];
-			die 'invalid signature length'
+			die_no_trace 'invalid signature length'
 				unless $validator->check($value);
 		},
 		version_status => {
@@ -579,7 +576,7 @@ my %types = (
 		},
 		key_deserializer => sub {
 			my $val = shift;
-			die 'invalid length'
+			die_no_trace 'invalid length'
 				unless length $val == 64;
 
 			my $leaf_hash = substr $val, 32, 32;
@@ -593,7 +590,7 @@ my %types = (
 		validator => sub {
 			my ($key, $value) = @_;
 			state $validator = ByteStrLen [64] | ByteStrLen [65];
-			die 'invalid signature length'
+			die_no_trace 'invalid signature length'
 				unless $validator->check($value);
 		},
 		version_status => {
@@ -669,7 +666,7 @@ my %types = (
 		validator => sub {
 			my ($value) = @_;
 			state $validator = ByteStrLen [32];
-			die 'invalid merkle root length'
+			die_no_trace 'invalid merkle root length'
 				unless $validator->check($value);
 		},
 		version_status => {
@@ -719,13 +716,13 @@ my %types = (
 
 	PSBT_OUT_AMOUNT => {
 		code => 0x03,
-		value_data => "Math::BigInt object",
+		value_data => "64 bit number",
 		serializer => sub {
 			state $sig = signature(positional => [SatoshiAmount]);
 			my $value = ($sig->(@_))[0];
-			return scalar reverse ensure_length $value->to_bytes, 8;
+			return encode_64bit($value);
 		},
-		deserializer => sub { Math::BigInt->from_bytes(scalar reverse shift) },
+		deserializer => sub { decode_64bit(shift) },
 		version_status => {
 			2 => REQUIRED,
 		},
@@ -761,7 +758,7 @@ my %types = (
 			my $action = sub {
 				my ($value, $depth) = @_;
 
-				die 'tree must have all its leaves unhashed'
+				die_no_trace 'tree must have all its leaves unhashed'
 					unless defined $value->{script};
 
 				my $serialized = $value->{script}->to_serialized;
@@ -865,7 +862,7 @@ foreach my $type (values %types) {
 }
 
 signature_for get_field_by_code => (
-	method => Str,
+	method => !!1,
 	positional => [PSBTMapType, PositiveOrZeroInt],
 );
 
@@ -890,7 +887,7 @@ sub get_field_by_code
 }
 
 signature_for get_field_by_name => (
-	method => Str,
+	method => !!1,
 	positional => [Str],
 );
 
@@ -906,7 +903,7 @@ sub get_field_by_name
 }
 
 signature_for get_fields_required_in_version => (
-	method => Str,
+	method => !!1,
 	positional => [PositiveOrZeroInt],
 );
 
@@ -925,7 +922,7 @@ sub get_fields_required_in_version
 }
 
 signature_for available_in_version => (
-	method => Object,
+	method => !!1,
 	positional => [PositiveOrZeroInt],
 );
 
@@ -937,7 +934,7 @@ sub available_in_version
 }
 
 signature_for required_in_version => (
-	method => Object,
+	method => !!1,
 	positional => [PositiveOrZeroInt],
 );
 
@@ -983,7 +980,7 @@ B<Required in the constructor.> Code of the field type defined in BIP174.
 
 B<Available in the constructor.> A map type this field belongs to. If not
 passed, it will be guessed from L</name>. Map types are defined as constants in
-C<Bitcoin::Crypto::Constants>.
+L<Bitcoin::Crypto::Constants>.
 
 =head3 serializer
 
@@ -1083,18 +1080,6 @@ Returns true if this field type is required in a given C<$version> number.
 
 Returns true if this field type is available in a given C<$version> number.
 
-=head1 EXCEPTIONS
-
-This module throws an instance of L<Bitcoin::Crypto::Exception> if it
-encounters an error. It can produce the following error types from the
-L<Bitcoin::Crypto::Exception> namespace:
-
-=over
-
-=item * PSBT - general error with the PSBT
-
-=back
-
 =head1 SEE ALSO
 
 =over
