@@ -1,10 +1,9 @@
 package Bitcoin::Crypto::Role::WithDerivationPath;
-$Bitcoin::Crypto::Role::WithDerivationPath::VERSION = '4.002';
-use v5.10;
-use strict;
+$Bitcoin::Crypto::Role::WithDerivationPath::VERSION = '4.003';
+use v5.14;
 use warnings;
 
-use Moo::Role;
+use Mooish::Base -standard, -role;
 
 requires qw(get_derivation_path);
 
